///////////////////////////////////////////////////////////////////////
//                                                                   //
//   Copyright 2012 David Alonso                                     //
//                                                                   //
//                                                                   //
// This file is part of CRIME.                                       //
//                                                                   //
// CRIME is free software: you can redistribute it and/or modify it  //
// under the terms of the GNU General Public License as published by //
// the Free Software Foundation, either version 3 of the License, or //
// (at your option) any later version.                               //
//                                                                   //
// CRIME is distributed in the hope that it will be useful, but      //
// WITHOUT ANY WARRANTY; without even the implied warranty of        //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU //
// General Public License for more details.                          //
//                                                                   //
// You should have received a copy of the GNU General Public License //
// along with CRIME.  If not, see <http://www.gnu.org/licenses/>.    //
//                                                                   //
///////////////////////////////////////////////////////////////////////
#include "common_gh.h"
#include <chealpix.h>
#include <fitsio.h>

static float *T_map_single;

static void write_single_fit(char *prefix,int i_nu,flouble *T_maps)
{
  char fname[256];
  long n_pix_ang=nside2npix(n_side);
  long npix_start=i_nu*n_pix_ang;

  sprintf(fname,"%s_%03d.fits",prefix,i_nu+1);

  long ii;
  for(ii=0;ii<n_pix_ang;ii++)
    T_map_single[ii]=(float)(T_maps[npix_start+ii]);
  //CBI can this casting be accelerated?

  he_write_healpix_map(&T_map_single,1,n_side,fname);
}

static void write_nu_table(char *prefix)
{
  int ii;
  char fname_nutable[256];
  FILE *ftab;
  double dnu=(nu_max-nu_min)/n_nu;

  sprintf(fname_nutable,"%s_nuTable.dat",prefix);
  ftab=fopen(fname_nutable,"w");
  if(ftab==NULL) error_open_file(fname_nutable);
  for(ii=0;ii<n_nu;ii++) {
    double nu0=nu_min+ii*dnu;
    double nuf=nu_min+(ii+1)*dnu;
    double z0=NU_21/nuf-1;
    double zf=NU_21/nu0-1;
    
    fprintf(ftab,"%d %lf %lf %lf %lf\n",ii+1,nu0,nuf,z0,zf);
  }
  fclose(ftab);
}

void write_maps(char *prefix,flouble *T_maps,flouble *T_maps_psources)
{
  int i_nu;
  write_nu_table(prefix);
  
  T_map_single=(float *)my_malloc(nside2npix(n_side)*sizeof(float));
  
  printf("*** Writing files %s_###.fits\n",prefix);
  for(i_nu=0;i_nu<n_nu;i_nu++)
    write_single_fit(prefix,i_nu,T_maps);
  
  if(do_psources) {
    char prefix_psources[256];
    sprintf(prefix_psources,"%s_ps",prefix);
    printf("*** Writing single slice files %s_###.fits\n",prefix_psources);
    for(i_nu=0;i_nu<n_nu;i_nu++)
      write_single_fit(prefix_psources,i_nu,T_maps_psources);
  }
  
  free(T_map_single);
}

void read_run_params(char *fname)
{
  //////
  // Reads and checks the parameter file
  FILE *fi;
  int n_lin,ii;
  
  //Read parameters from file
  printf("*** Reading run parameters \n");
  fi=fopen(fname,"r");
  if(fi==NULL) error_open_file(fname);
  n_lin=linecount(fi);
  rewind(fi);
  for(ii=0;ii<n_lin;ii++) {
    char s0[512],s1[64],s2[256];
    if(fgets(s0,sizeof(s0),fi)==NULL)
      error_read_line(fname,ii+1);
    if((s0[0]=='#')||(s0[0]=='\n')) continue;
    int sr=sscanf(s0,"%s %s",s1,s2);
    if(sr!=2)
      error_read_line(fname,ii+1);

    if(!strcmp(s1,"prefix_out="))
      sprintf(fnameOut,"%s",s2);
    else if(!strcmp(s1,"pk_filename="))
      sprintf(fnamePk,"%s",s2);
    else if(!strcmp(s1,"omega_M="))
      OmegaM=atof(s2);
    else if(!strcmp(s1,"omega_L="))
      OmegaL=atof(s2);
    else if(!strcmp(s1,"omega_B="))
      OmegaB=atof(s2);
    else if(!strcmp(s1,"h="))
      hhub=atof(s2);
    else if(!strcmp(s1,"w="))
      weos=atof(s2);
    else if(!strcmp(s1,"ns="))
      n_scal=atof(s2);
    else if(!strcmp(s1,"sigma_8="))
      sig8=atof(s2);
    else if(!strcmp(s1,"nu_min="))
      nu_min=atof(s2);
    else if(!strcmp(s1,"nu_max="))
      nu_max=atof(s2);
    else if(!strcmp(s1,"n_grid="))
      n_grid=atoi(s2);
    else if(!strcmp(s1,"n_side="))
      n_side=atoi(s2);
    else if(!strcmp(s1,"n_nu="))
      n_nu=atoi(s2);
    else if(!strcmp(s1,"seed="))
      signed_seed_rng=atoi(s2);
    else if(!strcmp(s1,"do_psources="))
      do_psources=atoi(s2);
    else
      fprintf(stderr,"CRIME: Unknown parameter %s\n",s1);
  }
  fclose(fi);

  cosmo_set();

  double dk=2*M_PI/l_box;
  double dtheta=sqrt(41253./(12*n_side*n_side));
  double drTmax=r_max*dtheta/RTOD;
  double drTmin=r_min*dtheta/RTOD;
  double dnu=(nu_max-nu_min)/n_nu;
  double drL=(r_max-r_min)/n_nu;
  double output_size=(flouble)(12*n_side*n_side*n_nu*sizeof(flouble))/(1024*1024*1024);
  printf("Run parameters: \n");
  printf("  %.3lf < nu/MHz < %.3lf\n",nu_min,nu_max);
  printf("  %.3lf < z < %.3lf\n",z_min,z_max);
  printf("  %.3lf < r/(Mpc/h) < %.3lf\n",r_min,r_max);
  printf("  L_box = %.3lf Mpc/h, N_grid = %d \n",l_box,n_grid);
  printf("  Scales resolved: %.3lE < k < %.3lE h/Mpc\n",dk,0.5*(n_grid-1)*dk);
  printf("  Fourier-space resolution: dk = %.3lE h/Mpc\n",dk);
  printf("  Real-space resolution: dx = %.3lE Mpc/h\n",l_box/n_grid);
  printf("  n_nu = %d, d_nu= %.3lf MHz, drL ~ %.3lf Mpc/h\n",
	 n_nu,dnu,drL);
  printf("  n_side = %ld, dtheta = %.3lf deg, %.3lf < drT/(Mpc/h) < %.3lf\n",
	 n_side,dtheta,drTmin,drTmax);
  printf("  Estimated output size ~ %.1lf GB \n",output_size);

  printf("\n");
}
